from nodeconductor.structure import perms as structure_perms


PERMISSION_LOGICS = (
    ('nodeconductor_sugarcrm.SugarCRMService', structure_perms.service_permission_logic),
    ('nodeconductor_sugarcrm.SugarCRMServiceProjectLink', structure_perms.service_project_link_permission_logic),
    ('nodeconductor_sugarcrm.CRM', structure_perms.resource_permission_logic),
)



from __future__ import unicode_literals

from django.db import models

from nodeconductor.core import utils as core_utils
from nodeconductor.cost_tracking.models import PayableMixin
from nodeconductor.quotas.fields import QuotaField, LimitAggregatorQuotaField, CounterQuotaField
from nodeconductor.quotas.models import QuotaModelMixin
from nodeconductor.structure import models as structure_models


class SugarCRMService(structure_models.Service):
    projects = models.ManyToManyField(
        structure_models.Project, related_name='sugarcrm_services', through='SugarCRMServiceProjectLink')

    class Meta:
        verbose_name = 'SugarCRM service'
        verbose_name_plural = 'SugarCRM services'
        unique_together = ('customer', 'settings')

    @classmethod
    def get_url_name(cls):
        return 'sugarcrm'


class SugarCRMServiceProjectLink(structure_models.ServiceProjectLink):
    service = models.ForeignKey(SugarCRMService)

    class Quotas(QuotaModelMixin.Quotas):
        user_limit_count = LimitAggregatorQuotaField(
            default_limit=50,
            get_children=lambda spl: CRM.objects.filter(service_project_link=spl),
            child_quota_name='user_count',
        )
        crm_count = CounterQuotaField(
            target_models=lambda: [CRM],
            path_to_scope='service_project_link',
            default_limit=-1,
        )

    class Meta(structure_models.ServiceProjectLink.Meta):
        verbose_name = 'SugarCRM service project link'
        verbose_name_plural = 'SugarCRM service project links'

    @classmethod
    def get_url_name(cls):
        return 'sugarcrm-spl'


class CRM(QuotaModelMixin, structure_models.PublishableResource, structure_models.ApplicationMixin,
          PayableMixin):
    service_project_link = models.ForeignKey(
        SugarCRMServiceProjectLink, related_name='crms', on_delete=models.PROTECT)

    api_url = models.CharField(max_length=127, help_text='CRMs OpenStack instance access URL.')
    admin_username = models.CharField(max_length=60)
    admin_password = models.CharField(max_length=255)
    instance_url = models.URLField(blank=True, help_text='CRMs OpenStack instance URL in NC.')

    class Quotas(QuotaModelMixin.Quotas):
        user_count = QuotaField(default_limit=0)

    class Meta:
        verbose_name = 'CRM'
        verbose_name_plural = 'CRMs'

    @classmethod
    def get_url_name(cls):
        return 'sugarcrm-crms'

    @property
    def full_name(self):
        return 'SugarCRM Instance %s' % self.name

    def get_backend(self):
        from .backend import SugarCRMBackend
        return SugarCRMBackend(settings=self.service_project_link.service.settings, crm=self)

    def get_instance(self):
        """ Restore instance from URL """
        return core_utils.instance_from_url(self.instance_url)



from . import views


def register_in(router):
    router.register(r'sugarcrm', views.SugarCRMServiceViewSet, base_name='sugarcrm')
    router.register(r'sugarcrm-crms', views.CRMViewSet, base_name='sugarcrm-crms')
    router.register(r'sugarcrm-service-project-link', views.SugarCRMServiceProjectLinkViewSet, base_name='sugarcrm-spl')
    router.register(r'sugarcrm-crms/(?P<crm_uuid>[\w]+)/users', views.CRMUserViewSet, base_name='sugarcrm-users')



from django.dispatch import Signal

# SugarCRM user signals
user_post_save = Signal(providing_args=['old_user', 'new_user', 'crm', 'created'])
user_post_delete = Signal(providing_args=['user', 'crm'])



import logging
import md5
import urlparse

from django.core.exceptions import ObjectDoesNotExist
from django.core.urlresolvers import Resolver404
from django.utils import six
import requests
from rest_framework.reverse import reverse
import sugarcrm

from nodeconductor.core.tasks import send_task
from nodeconductor.core.utils import pwgen
from nodeconductor.structure import ServiceBackend, ServiceBackendError

from . import models


logger = logging.getLogger(__name__)


class SugarCRMBackendError(ServiceBackendError):
    pass


class SugarCRMBaseBackend(ServiceBackend):

    def __init__(self, settings, crm=None):
        self.settings = settings
        self.crm = crm

    def provision(self, crm, user_count):
        crm.set_quota_limit(crm.Quotas.user_count, user_count)
        send_task('sugarcrm', 'provision_crm')(
            crm.uuid.hex,
        )

    def destroy(self, crm, force=False):
        # CRM cannot be stopped by user - so we need to stop it before deletion on destroy
        crm.state = crm.States.STOPPING_SCHEDULED
        crm.save()
        send_task('sugarcrm', 'stop_and_destroy_crm')(
            crm.uuid.hex,
            force=force,
        )

    def sync(self):
        pass


class SugarCRMBackend(SugarCRMBaseBackend):
    """ Sugar CRM backend methods

    Backend uses two clients for operations:
    1. NodeCondcutor OpenStack client uses to NC OpenStack application endpoints for operations with SugarCRM VM.
    2. SugarCRM client interacts with SugarCRM API.
    """

    DEFAULTS = {
        'user_data': "#cloud-config:\nruncmd:\n  - [ bootstrap, -p, {password}, -k, {license_code}, -v]",
        'protocol': "http",
    }

    CRM_ADMIN_NAME = 'admin'

    class NodeConductorOpenStackClient(object):

        def __init__(self, template_url, username, password):
            self.credentials = {
                'username': username,
                'password': password,
            }
            parsed = urlparse.urlparse(template_url)
            self.scheme = parsed.scheme
            self.netloc = parsed.netloc

        def authenticate(self):
            url = self._prepare_url(reverse('auth-password'))
            response = requests.post(url, data=self.credentials, verify=False)
            if response.ok:
                self.token = response.json()['token']
            else:
                raise SugarCRMBackendError('Cannot authenticate as %s' % self.credentials['username'])

        def _prepare_url(self, url):
            """ Add scheme, host and port to url """
            if not url.startswith('http'):
                return urlparse.urlunsplit([self.scheme, self.netloc, url, '', ''])
            return url

        def _make_request(self, method, url, retry_if_authentication_fails=True, **kwargs):
            if not hasattr(self, 'token'):
                self.authenticate()
            headers = kwargs.get('headers', {})
            headers['Authorization'] = 'Token %s' % self.token
            kwargs['headers'] = headers
            kwargs['verify'] = kwargs.get('verify', False)

            url = self._prepare_url(url)
            response = getattr(requests, method)(url, **kwargs)
            if response.status_code == requests.status_codes.codes.unauthorized and retry_if_authentication_fails:
                return self._make_request(method, url, retry_if_authentication_fails=False, **kwargs)
            else:
                return response

        def get(self, url, **kwargs):
            return self._make_request('get', url, **kwargs)

        def post(self, url, **kwargs):
            return self._make_request('post', url, **kwargs)

        def delete(self, url, **kwargs):
            return self._make_request('delete', url, **kwargs)

    class SugarCRMClient(object):

        class UserStatuses(object):
            ACTIVE = 'Active'
            INACTIVE = 'Inactive'
            RESERVED = 'Reserved'

        def __init__(self, url, username, password):
            self.v4_url = url + '/service/v4/rest.php'
            self.v10_url = url + '/rest/v10/'
            self.username = username
            self.password = password
            self.v4_session = sugarcrm.Session(self.v4_url, username, password)

        def execute_v10_request(self, method, url, json_data):
            method = getattr(requests, method.lower())
            if not hasattr(self, '_v10_headers'):
                self._v10_headers = self._get_v10_headers()
            return method(url, json=json_data, headers=self._v10_headers)

        def _get_v10_headers(self):
            auth_url = self.v10_url + 'oauth2/token/'
            json_data = {
                'client_id': 'sugar',
                'client_secret': '',
                'grant_type': 'password',
                'password': self.password,
                'platform': 'base',
                'username': self.username,
            }
            response = requests.post(auth_url, json=json_data).json()
            return {'oauth-token': response['access_token']}

        def create_user(self, **kwargs):
            user = sugarcrm.User()
            for key, value in kwargs.items():
                setattr(user, key, value)
            return self.v4_session.set_entry(user)

        def update_user(self, user, **kwargs):
            # It is possible to update user status only with v10 API
            if 'status' in kwargs:
                url = self.v10_url + 'Users/%s/' % user.id
                self.execute_v10_request('PUT', url, json_data={'status': kwargs['status']})
            for key, value in kwargs.items():
                setattr(user, key, value)
            return self.v4_session.set_entry(user)

        def get_user(self, user_id):
            return self.v4_session.get_entry('Users', user_id)

        def list_users(self, **kwargs):
            # admin users should not be visible
            user_query = sugarcrm.User(is_admin='0')
            user_count = self.v4_session.get_entries_count(user_query)

            step = 100
            users = []
            for offset in range(0, user_count, step):
                users += self.v4_session.get_entry_list(user_query, max_results=step, offset=offset)
            # do not show users that are reserved by sugarcrm:
            users = [user for user in users if user.status != self.UserStatuses.RESERVED]
            # XXX: SugarCRM cannot filter 2 arguments together - its easier to filter users here.
            return [user for user in users if all(getattr(user, k) == v for k, v in kwargs.items())]

        def delete_user(self, user):
            user.deleted = 1
            self.v4_session.set_entry(user)

    def __init__(self, *args, **kwargs):
        super(SugarCRMBackend, self).__init__(*args, **kwargs)

        self.template_url = self.settings.backend_url
        self.nc_client = self.NodeConductorOpenStackClient(
            self.template_url, self.settings.username, self.settings.password)

    @property
    def sugar_client(self):
        if hasattr(self, '_sugar_client'):
            return self._sugar_client
        if self.crm is None:
            raise SugarCRMBackendError('It is impossible to use sugar client if CRM is not specified for backend.')
        try:
            self._sugar_client = self.SugarCRMClient(self.crm.api_url, self.crm.admin_username, self.crm.admin_password)
        except KeyError:
            raise SugarCRMBackendError('Cannot connect to CRM backend.')
        return self._sugar_client

    def schedule_crm_instance_provision(self, crm):
        # prepare data for template group provisioning
        user_data = self.settings.get_option('user_data')
        admin_username = self.CRM_ADMIN_NAME
        admin_password = pwgen()
        user_data = user_data.format(
            password=admin_password, license_code=self.settings.get_option('license_code'))

        template_data = [{
            'name': crm.name,
            'user_data': user_data,
        }]
        # schedule template provisioning
        response = self.nc_client.post(self.template_url + 'provision/', json=template_data)
        if not response.ok:
            raise SugarCRMBackendError(
                'Cannot provision OpenStack instance for CRM "%s": response code - %s, response content: %s.'
                'Request URL: %s, request body: %s' %
                (crm.name, response.status_code, response.content, response.request.url, response.request.body))
        # store template group result as backend_id
        crm.admin_password = admin_password
        crm.admin_username = admin_username
        crm.backend_id = response.json()['url']
        crm.instance_url = response.json()['provisioned_resources']['OpenStack.Instance']
        crm.save()

        logger.info('Successfully scheduled instance provision for CRM "%s"', crm.name)

    def schedule_crm_instance_stopping(self, crm):
        response = self.nc_client.post(crm.instance_url + 'stop/')
        if not response.ok:
            raise SugarCRMBackendError(
                'Cannot stop OpenStack instance for CRM "%s": response code - %s, response content: %s.'
                'Request URL: %s, request body: %s' %
                (crm.name, response.status_code, response.content, response.request.url, response.request.body))

        logger.info('Successfully scheduled instance stopping for CRM "%s"', crm.name)

    def schedule_crm_instance_deletion(self, crm):
        response = self.nc_client.delete(crm.instance_url)
        if not response.ok:
            raise SugarCRMBackendError(
                'Cannot delete OpenStack instance for CRM "%s": response code - %s, response content: %s.'
                'Request URL: %s, request body: %s' %
                (crm.name, response.status_code, response.content, response.request.url, response.request.body))

        logger.info('Successfully scheduled instance deletion for CRM "%s"', crm.name)

    def pull_crm_sla(self, crm):
        """ Copy OpenStack instance SLA and events as CRM events """
        try:
            instance = crm.get_instance()
        except (Resolver404, ObjectDoesNotExist) as e:
            crm.error_message = 'Cannot get instance for CRM %s (PK: %s). Error: %s' % (crm.name, crm.pk, e)
            crm.set_erred()
            crm.save()
            logger.error(crm.error_message)
            six.reraise(SugarCRMBackendError, e)
        else:
            for item in instance.sla_items.all():
                crm.sla_items.update_or_create(
                    period=item.period,
                    defaults={'value': item.value, 'agreed_value': item.agreed_value},
                )
            for state_transition in instance.state_items.all():
                crm.state_items.update_or_create(
                    period=state_transition.period,
                    timestamp=state_transition.timestamp,
                    defaults={'state': state_transition.state},
                )

        logger.info('Successfully pulled SLA for CRM "%s"', crm.name)

    def get_crm_instance_details(self, crm):
        """ Get details of instance that corresponds given CRM """
        response = self.nc_client.get(crm.instance_url)
        if not response.ok:
            raise SugarCRMBackendError(
                'Cannot get details of CRMs instance: response code - %s, response content: %s.'
                'Request URL: %s' % (response.status_code, response.content, response.request.url))
        return response.json()

    def get_crm_template_group_result_details(self, crm):
        """ Get details of CRMs template group provision result """
        if not crm.backend_id:
            raise SugarCRMBackendError('Cannot get instance URL for CRM without backend id')
        template_group_result_url = crm.backend_id
        response = self.nc_client.get(template_group_result_url)
        if not response.ok:
            raise SugarCRMBackendError(
                'Cannot get CRMs provision result details: response code - %s, response content: %s.'
                'Request URL: %s' % (response.status_code, response.content, response.request.url))
        return response.json()

    def create_user(self, user_name, password, last_name, **kwargs):
        encoded_password = self._encode_password(password)
        try:
            user = self.sugar_client.create_user(
                user_name=user_name, user_hash=encoded_password, last_name=last_name, **kwargs)
        except (requests.exceptions.RequestException, sugarcrm.SugarError) as e:
            raise SugarCRMBackendError(
                'Cannot create user %s on CRM "%s". Error: %s' % (user_name, self.crm.name, e))

        self.crm.add_quota_usage(self.crm.Quotas.user_count, 1)
        logger.info('Successfully created user "%s" for CRM "%s"', user_name, self.crm.name)
        return user

    def update_user(self, user, **kwargs):
        if 'password' in kwargs:
            kwargs['user_hash'] = self._encode_password(kwargs.pop('password'))
        try:
            user = self.sugar_client.update_user(user, **kwargs)
        except (requests.exceptions.RequestException, sugarcrm.SugarError) as e:
            raise SugarCRMBackendError(
                'Cannot update user %s on CRM "%s". Error: %s' % (user.user_name, self.crm.name, e))

        logger.info('Successfully updated user "%s" for CRM "%s"', user.user_name, self.crm.name)
        return user

    def _encode_password(self, password):
        return md5.new(password).hexdigest()

    def delete_user(self, user):
        try:
            self.sugar_client.delete_user(user)
        except (requests.exceptions.RequestException, sugarcrm.SugarError) as e:
            raise SugarCRMBackendError(
                'Cannot delete user with id %s from CRM "%s". Error: %s' % (user.id, self.crm.name, e))

        self.crm.add_quota_usage(self.crm.Quotas.user_count, -1)
        logger.info('Successfully deleted user with id %s on CRM "%s"', user.id, self.crm.name)

    def get_user(self, user_id):
        try:
            return self.sugar_client.get_user(user_id)
        except (requests.exceptions.RequestException, sugarcrm.SugarError) as e:
            raise SugarCRMBackendError(
                'Cannot get user with id %s from CRM "%s". Error: %s' % (user_id, self.crm.name, e))

    def list_users(self, **kwargs):
        try:
            return self.sugar_client.list_users(**kwargs)
        except (requests.exceptions.RequestException, sugarcrm.SugarError) as e:
            raise SugarCRMBackendError('Cannot get users from CRM "%s". Error: %s' % (self.crm.name, e))

    def sync_user_quota(self):
        """ Sync CRM quotas with backend """
        active_users = self.list_users(status=self.SugarCRMClient.UserStatuses.ACTIVE)
        self.crm.set_quota_usage(self.crm.Quotas.user_count, len(active_users))

    def get_stats(self):
        links = models.CRM.objects.filter(
            service_project_link__service__settings=self.settings)
        quota_values = models.CRM.get_sum_of_quotas_as_dict(links, quota_names=('user_count',))
        stats = {
            'user_count_quota': quota_values.get('user_count', -1.0),
            'user_count_usage': quota_values.get('user_count_usage', 0.0)
        }

        try:
            quota = self.settings.quotas.get(name='sugarcrm_user_count')
            stats['user_count'] = quota.limit
        except ObjectDoesNotExist:
            stats['user_count'] = -1.0
        return stats



import logging
import sys

from celery import shared_task, chain
from django.utils import six, timezone

from nodeconductor.core import utils as core_utils
from nodeconductor.core.tasks import save_error_message, transition, retry_if_false, BackendMethodTask

from .backend import SugarCRMBackendError
from .models import CRM


logger = logging.getLogger(__name__)


@shared_task(name='nodeconductor.sugarcrm.provision_crm')
def provision_crm(crm_uuid):
    chain(
        schedule_crm_instance_provision.si(crm_uuid),
        wait_for_crm_template_group_provision.si(crm_uuid),
        init_crm_api_url.si(crm_uuid),
        init_crm_quotas.si(crm_uuid),
    ).apply_async(
        link=set_online.si(crm_uuid),
        link_error=set_erred.si(crm_uuid)
    )


@shared_task(name='nodeconductor.sugarcrm.stop_and_destroy_crm')
def stop_and_destroy_crm(crm_uuid, force=False):
    if not force:
        error_callback = set_erred.si(crm_uuid)
    else:
        error_callback = force_delete.si(crm_uuid)
    chain(
        schedule_crm_instance_stopping.si(crm_uuid),
        wait_for_crm_instance_state.si(crm_uuid, state='Offline'),
        set_offline.si(crm_uuid),
        schedule_deletion.si(crm_uuid),
        schedule_crm_instance_deletion.si(crm_uuid),
    ).apply_async(
        link=delete.si(crm_uuid),
        link_error=error_callback,
    )


@shared_task
@transition(CRM, 'begin_provisioning')
@save_error_message
def schedule_crm_instance_provision(crm_uuid, transition_entity=None):
    crm = transition_entity
    backend = crm.get_backend()
    backend.schedule_crm_instance_provision(crm)


@shared_task
@transition(CRM, 'begin_stopping')
@save_error_message
def schedule_crm_instance_stopping(crm_uuid, transition_entity=None):
    crm = transition_entity
    backend = crm.get_backend()
    backend.schedule_crm_instance_stopping(crm)


@shared_task
@transition(CRM, 'begin_deleting')
@save_error_message
def schedule_crm_instance_deletion(crm_uuid, transition_entity=None):
    crm = transition_entity
    backend = crm.get_backend()
    backend.schedule_crm_instance_deletion(crm)


@shared_task(max_retries=120, default_retry_delay=20)
@retry_if_false
def wait_for_crm_instance_state(crm_uuid, state, erred_state='Erred'):
    crm = CRM.objects.get(uuid=crm_uuid)
    backend = crm.get_backend()
    details = backend.get_crm_instance_details(crm)
    current_state = details['state']
    logger.info('Checking state for CRM "%s" (UUID: %s) instance. Current value: %s.',
                crm.name, crm.uuid.hex, current_state)
    if current_state == erred_state:
        message = ('CRM "%s" (UUID: %s) instance with UUID %s become erred. Check OpenStack app logs '
                   'for more details.' % (crm.name, crm.uuid.hex, details['uuid']))
        crm.error_message = message
        crm.save()
        raise SugarCRMBackendError(message)
    return current_state == state


@shared_task(max_retries=120, default_retry_delay=20)
@retry_if_false
def wait_for_crm_template_group_provision(crm_uuid):
    crm = CRM.objects.get(uuid=crm_uuid)
    backend = crm.get_backend()
    details = backend.get_crm_template_group_result_details(crm)
    logger.info('Checking state of CRM "%s" (UUID: %s) provision result. Current state message: %s.',
                crm.name, crm.uuid.hex, details['state_message'])
    if details['is_erred']:
        message = ('CRM "%s" (UUID: %s) provision failed. result URL %s, message %s' % (
                   crm.name, crm.uuid.hex, details['url'], details['error_message']))
        crm.error_message = message
        crm.save()
        raise SugarCRMBackendError(message)
    return details['is_finished']


@shared_task
def init_crm_api_url(crm_uuid):
    """ Init CRM API URL """
    crm = CRM.objects.get(uuid=crm_uuid)
    settings = crm.service_project_link.service.settings
    backend = crm.get_backend()
    external_ips = backend.get_crm_instance_details(crm)['external_ips']
    if not external_ips:
        raise SugarCRMBackendError(
            'Cannot use OpenStack instance with name "%s" for CRM - it does not have external IP.' % crm.name)
    crm.api_url = '{protocol}://{external_ip}'.format(
        protocol=settings.get_option('protocol'),
        external_ip=external_ips[0])
    # we consider CRM as activated at this point
    crm.start_time = timezone.now()
    crm.save()


@shared_task(max_retries=30, default_retry_delay=10)
@retry_if_false
def init_crm_quotas(crm_uuid):
    """ Init CRM quotas """
    crm = CRM.objects.get(uuid=crm_uuid)
    backend = crm.get_backend()
    # It can take some time to initialize SugarCRM API, so we need to poll its API
    try:
        backend.sync_user_quota()
    except SugarCRMBackendError:
        return False
    else:
        return True


@shared_task
@transition(CRM, 'set_online')
def set_online(crm_uuid, transition_entity=None):
    pass


@shared_task
@transition(CRM, 'set_offline')
def set_offline(crm_uuid, transition_entity=None):
    pass


@shared_task
@transition(CRM, 'schedule_deletion')
def schedule_deletion(crm_uuid, transition_entity=None):
    pass


@shared_task
@transition(CRM, 'set_erred')
def set_erred(crm_uuid, transition_entity=None):
    pass


@shared_task
def delete(crm_uuid):
    CRM.objects.get(uuid=crm_uuid).delete()


@shared_task
def force_delete(crm_uuid):
    """ Schedule corresponding OpenStack instance deletion """
    crm = CRM.objects.get(uuid=crm_uuid)
    backend = crm.get_backend()
    try:
        backend.schedule_crm_instance_deletion(crm)
    except:
        six.reraise(*sys.exc_info())
    finally:
        crm.delete()


# celerybeat tasks:

@shared_task(name='nodeconductor.sugarcrm.sync_crms_quotas')
def sync_crms_quotas():
    """ Update quota usage from backend for all CRMs """
    for crm in CRM.objects.filter(state=CRM.States.ONLINE):
        sync_crm_quotas.delay(crm.uuid.hex)


@shared_task
def sync_crm_quotas(crm_uuid):
    crm = CRM.objects.get(uuid=crm_uuid)
    backend = crm.get_backend()
    backend.sync_user_quota()


@shared_task(name='nodeconductor.sugarcrm.pull_sla')
def pull_sla():
    """ Copy OpenStack instance SLA and events as CRM events """
    for crm in CRM.objects.filter(state=CRM.States.ONLINE):
        BackendMethodTask().delay(core_utils.serialize_instance(crm), 'pull_crm_sla')



from nodeconductor.core import NodeConductorExtension


class SugarCRMExtension(NodeConductorExtension):

    @staticmethod
    def django_app():
        return 'nodeconductor_sugarcrm'

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in

    @staticmethod
    def celery_tasks():
        from datetime import timedelta
        return {
            'sugarcrm-sync-crms-quotas': {
                'task': 'nodeconductor.sugarcrm.sync_crms_quotas',
                'schedule': timedelta(days=1),
                'args': ()
            },
            'sugarcrm-pull-sla': {
                'task': 'nodeconductor.sugarcrm.pull_sla',
                'schedule': timedelta(minutes=5),
            },
        }



from django.contrib.contenttypes.models import ContentType

from nodeconductor.cost_tracking import CostTrackingBackend
from nodeconductor.cost_tracking.models import DefaultPriceListItem

from .models import CRM


USERS = 'users'
USERS_KEY = 'count'
SUPPORT = 'support'
SUPPORT_KEY = 'premium'


class SugarCRMCostTrackingBackend(CostTrackingBackend):
    NUMERICAL = [USERS]

    @classmethod
    def get_default_price_list_items(cls):
        content_type = ContentType.objects.get_for_model(CRM)

        yield DefaultPriceListItem(
            item_type=USERS, key=USERS_KEY,
            resource_content_type=content_type)

        yield DefaultPriceListItem(
            item_type=SUPPORT, key=SUPPORT_KEY,
            resource_content_type=content_type)

    @classmethod
    def get_used_items(cls, resource):
        user_count = resource.quotas.get(name=resource.Quotas.user_count).limit
        return [(USERS, USERS_KEY, user_count)]



import logging

from django.core.mail import send_mail


logger = logging.getLogger(__name__)


def sms_user_password(crm, phone, password):
    settings = crm.service_project_link.service.settings
    sender = settings.get_option('sms_email_from')
    recipient = settings.get_option('sms_email_rcpt')

    if sender and recipient and '{phone}' in recipient:
        send_mail(
            '', 'Your OTP is: %s' % password, sender,
            [recipient.format(phone=phone)], fail_silently=True)
    else:
        logger.warning('SMS was not sent to SugarCRM user, because `sms_email_from` and `sms_email_rcpt` '
                       'were not configured properly.')



import copy

from django.shortcuts import get_object_or_404
from django.utils.crypto import get_random_string
from rest_framework import status, viewsets, exceptions
from rest_framework.decorators import detail_route
from rest_framework.response import Response

from nodeconductor.structure import views as structure_views
from nodeconductor.structure.managers import filter_queryset_for_user
from . import models, serializers, backend, signals, utils, log

event_logger = log.event_logger


class SugarCRMServiceViewSet(structure_views.BaseServiceViewSet):
    queryset = models.SugarCRMService.objects.all()
    serializer_class = serializers.ServiceSerializer


class SugarCRMServiceProjectLinkViewSet(structure_views.BaseServiceProjectLinkViewSet):
    queryset = models.SugarCRMServiceProjectLink.objects.all()
    serializer_class = serializers.ServiceProjectLinkSerializer


class CRMViewSet(structure_views.BaseOnlineResourceViewSet):
    queryset = models.CRM.objects.all()
    serializer_class = serializers.CRMSerializer

    def perform_provision(self, serializer):
        resource = serializer.save()
        backend = resource.get_backend()
        backend.provision(resource, user_count=serializer.validated_data['user_count'])


class CRMNotOnline(exceptions.APIException):
    status_code = 409


class CRMUserViewSet(viewsets.ViewSet):

    def initial(self, request, crm_uuid, *args, **kwargs):
        super(CRMUserViewSet, self).initial(request, crm_uuid, *args, **kwargs)
        queryset = filter_queryset_for_user(models.CRM.objects.all(), request.user)
        self.crm = get_object_or_404(queryset, uuid=crm_uuid)
        if self.crm.state != models.CRM.States.ONLINE:
            raise CRMNotOnline('Cannot execute user-related operations with CRM if it is not ONLINE')
        self.backend = self.crm.get_backend()

    def handle_exception(self, exc):
        if isinstance(exc, backend.SugarCRMBackendError):
            self.crm.set_erred()
            self.crm.error_message = str(exc)
            self.crm.save()
            return super(CRMUserViewSet, self).handle_exception(
                CRMNotOnline('Error appeared during request to CRM API.'))
        return super(CRMUserViewSet, self).handle_exception(exc)

    def get_serializer_context(self):
        return {'crm': self.crm, 'request': self.request, 'view': self}

    # XXX: This method should be replaced by default filter then CRM users will be moved to separate model
    def get_filtered_users(self, request):
        supported_filters = ['first_name', 'last_name', 'user_name', 'status']
        filter_kwargs = {f: request.query_params[f] for f in supported_filters if f in request.query_params}
        return self.backend.list_users(**filter_kwargs)

    def list(self, request, crm_uuid):
        users = self.get_filtered_users(request)
        serializer = serializers.CRMUserSerializer(users, many=True, context=self.get_serializer_context())
        return Response(serializer.data)

    def retrieve(self, request, crm_uuid, pk=None):
        user = self.backend.get_user(pk)
        if user is None or int(user.is_admin):
            return Response(status=status.HTTP_404_NOT_FOUND)
        serializer = serializers.CRMUserSerializer(user, context=self.get_serializer_context())
        return Response(serializer.data)

    def destroy(self, request, crm_uuid, pk=None):
        user = self.backend.get_user(pk)
        if user is None or int(user.is_admin):
            return Response(status=status.HTTP_404_NOT_FOUND)
        self.backend.delete_user(user)
        signals.user_post_delete.send(sender=models.CRM, user=user, crm=self.crm)
        return Response(status=status.HTTP_204_NO_CONTENT)

    def create(self, request, crm_uuid):
        serializer = serializers.CRMUserSerializer(data=request.data, context=self.get_serializer_context())
        serializer.is_valid(raise_exception=True)

        notify = serializer.validated_data.pop('notify', False)

        password = get_random_string(10)
        data = dict(password=password, **serializer.validated_data)
        user = self.backend.create_user(status='Active', **data)
        user_data = serializers.CRMUserSerializer(user, context=self.get_serializer_context()).data
        signals.user_post_save.send(sender=models.CRM, old_user=None, new_user=user, crm=self.crm, created=True)

        if notify:
            utils.sms_user_password(self.crm, user.phone_mobile, password)

        return Response(dict(password=password, **user_data), status=status.HTTP_201_CREATED)

    def update(self, request, crm_uuid, pk=None):
        return self.partial_update(request, crm_uuid, pk=pk)

    def partial_update(self, request, crm_uuid, pk=None):
        old_user = self.backend.get_user(pk)
        if old_user is None or int(old_user.is_admin):
            return Response(status=status.HTTP_404_NOT_FOUND)
        serializer = serializers.CRMUserSerializer(
            data=request.data, instance=old_user, context=self.get_serializer_context())
        serializer.is_valid(raise_exception=True)

        new_user = self.backend.update_user(copy.deepcopy(old_user), **serializer.validated_data)
        user_data = serializers.CRMUserSerializer(new_user, context=self.get_serializer_context()).data
        signals.user_post_save.send(
            sender=models.CRM, old_user=old_user, new_user=new_user, crm=self.crm, created=False)
        return Response(user_data, status.HTTP_200_OK)

    @detail_route(methods=['post'])
    def password(self, request, crm_uuid, pk=None):
        user = self.backend.get_user(pk)
        serializer = serializers.UserPasswordSerializer(data=request.data, context={'user': user})
        serializer.is_valid(raise_exception=True)

        password = get_random_string(length=10)
        self.backend.update_user(user, password=password)

        if serializer.validated_data.get('notify'):
            utils.sms_user_password(self.crm, user.phone_mobile, password)

        event_logger.sugarcrm_user.info(
            'SugarCRM user {user_name} password has been reset.',
            event_type='sugarcrm_user_password_reset',
            event_context={
                'user_name': user.user_name,
                'crm': self.crm
            })

        return Response({'password': password}, status=status.HTTP_200_OK)



from django.core.management.base import BaseCommand

from ...models import CRM


class Command(BaseCommand):
    help_text = "Initialize CRMs instance_url field."

    def handle(self, *args, **options):
        for crm in CRM.objects.all():
            try:
                b = crm.get_backend()
                crm.instance_url = b.get_crm_template_group_result_details(crm)['provisioned_resources']['OpenStack.Instance']
                crm.save()
            except Exception as e:
                self.stdout.write('Cannot initialize isntance_url for CRM: %s (UUID: %s). Error: %s' % (
                    crm.name, crm.uuid.hex, e))
            else:
                self.stdout.write('Successfully initialized instance url for CRM: %s (UUID: %s).' % (
                    crm.name, crm.uuid.hex))



from nodeconductor.logging.loggers import EventLogger, event_logger

from .models import CRM


class SugarCRMUserEventLogger(EventLogger):
    user_name = basestring
    crm = CRM

    class Meta:
        event_types = (
            'sugarcrm_user_creation_succeeded',
            'sugarcrm_user_update_succeeded',
            'sugarcrm_user_deletion_succeeded',
            'sugarcrm_user_activation_succeeded',
            'sugarcrm_user_deactivation_succeeded',
            'sugarcrm_user_password_reset',
        )


event_logger.register('sugarcrm_user', SugarCRMUserEventLogger)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('nodeconductor_sugarcrm', '0010_crm_instance_url'),
    ]

    operations = [
        migrations.AddField(
            model_name='crm',
            name='publishing_state',
            field=models.CharField(default='not published', max_length=30, choices=[('not published', 'Not published'), ('published', 'Published'), ('requested', 'Requested')]),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('nodeconductor_sugarcrm', '0011_crm_publishing_state'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='sugarcrmserviceprojectlink',
            name='error_message',
        ),
        migrations.RemoveField(
            model_name='sugarcrmserviceprojectlink',
            name='state',
        ),
    ]



from django import forms

from rest_framework import serializers

from nodeconductor.structure import models as structure_models
from nodeconductor.template.forms import ResourceTemplateForm
from nodeconductor.template.serializers import BaseResourceTemplateSerializer
from nodeconductor_sugarcrm import models


# TODO: Filter service settings of defined type
class CRMProvisionTemplateForm(ResourceTemplateForm):
    service_settings = forms.ModelChoiceField(
        label="SugarCRM service settings",
        queryset=structure_models.ServiceSettings.objects.all(),
        required=False)

    user_count = forms.IntegerField(required=False)

    class Meta(ResourceTemplateForm.Meta):
        fields = ResourceTemplateForm.Meta.fields + ('service_settings', 'user_count')

    class Serializer(BaseResourceTemplateSerializer):
        service_settings = serializers.HyperlinkedRelatedField(
            view_name='servicesettings-detail',
            queryset=structure_models.ServiceSettings.objects.all(),
            lookup_field='uuid',
            required=False,
        )
        user_count = serializers.IntegerField(required=False)

    @classmethod
    def get_serializer_class(cls):
        return cls.Serializer

    @classmethod
    def get_model(cls):
        return models.CRM



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('nodeconductor_sugarcrm', '0009_remove_crm_size_field'),
    ]

    operations = [
        migrations.AddField(
            model_name='crm',
            name='instance_url',
            field=models.URLField(help_text='CRMs OpenStack instance URL in NC.', blank=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='crm',
            name='api_url',
            field=models.CharField(help_text='CRMs OpenStack instance access URL.', max_length=127),
            preserve_default=True,
        ),
    ]



from django.apps import AppConfig


class SugarCRMConfig(AppConfig):
    name = 'nodeconductor_sugarcrm'
    verbose_name = "NodeConductor SugarCRM"
    service_name = 'SugarCRM'

    def ready(self):
        from nodeconductor.cost_tracking import CostTrackingRegister
        from nodeconductor.quotas.fields import LimitAggregatorQuotaField
        from nodeconductor.structure import SupportedServices
        from nodeconductor.template import TemplateRegistry

        # structure
        from .backend import SugarCRMBackend
        SupportedServices.register_backend(SugarCRMBackend)

        # cost tracking
        from .cost_tracking import SugarCRMCostTrackingBackend
        CostTrackingRegister.register(self.label, SugarCRMCostTrackingBackend)

        # template
        from .template import CRMProvisionTemplateForm
        TemplateRegistry.register(CRMProvisionTemplateForm)

        from nodeconductor.structure.models import ServiceSettings
        from . import handlers, signals as sugarcrm_signals
        CRM = self.get_model('CRM')
        SugarCRMServiceProjectLink = self.get_model('SugarCRMServiceProjectLink')

        sugarcrm_signals.user_post_save.connect(
            handlers.log_user_post_save,
            sender=CRM,
            dispatch_uid='nodeconductor_sugarcrm.handlers.log_user_post_save'
        )

        sugarcrm_signals.user_post_delete.connect(
            handlers.log_user_post_delete,
            sender=CRM,
            dispatch_uid='nodeconductor_sugarcrm.handlers.log_user_post_delete'
        )

        ServiceSettings.add_quota_field(
            name='sugarcrm_user_count',
            quota_field=LimitAggregatorQuotaField(
                creation_condition=lambda service_settings: service_settings.type == SugarCRMConfig.service_name,
                get_children=lambda service_settings: SugarCRMServiceProjectLink.objects.filter(
                                                        service__settings=service_settings),
                child_quota_name='user_limit_count',
            ),
        )



from . import log


event_logger = log.event_logger


# logging
def log_user_post_save(sender, old_user, new_user, crm, created=False, **kwargs):
    if created:
        event_logger.sugarcrm_user.info(
            'User {user_name} has been created in CRM {crm_name}.',
            event_type='sugarcrm_user_creation_succeeded',
            event_context={
                'user_name': new_user.user_name,
                'crm': crm,
            })
    else:
        # check if activation has been requested
        if old_user.status != new_user.status:
            if new_user.status == 'Active':
                event_logger.sugarcrm_user.info(
                    'User {user_name} has been activated in CRM {crm_name}.',
                    event_type='sugarcrm_user_activation_succeeded',
                    event_context={
                        'user_name': new_user.user_name,
                        'crm': crm,
                    })
            elif new_user.status == 'Inactive':
                event_logger.sugarcrm_user.info(
                    'User {user_name} has been deactivated in CRM {crm_name}.',
                    event_type='sugarcrm_user_deactivation_succeeded',
                    event_context={
                        'user_name': new_user.user_name,
                        'crm': crm,
                    })

        event_logger.sugarcrm_user.info(
            'User {user_name} has been updated in CRM {crm_name}.',
            event_type='sugarcrm_user_update_succeeded',
            event_context={
                'user_name': new_user.user_name,
                'crm': crm,
            })


def log_user_post_delete(sender, user, crm, **kwargs):
    event_logger.sugarcrm_user.info(
        'User {user_name} has been deleted from CRM {crm_name}.',
        event_type='sugarcrm_user_deletion_succeeded',
        event_context={
            'user_name': user.user_name,
            'crm': crm,
        })



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import taggit.managers
import django_fsm
import nodeconductor.core.models
import django.db.models.deletion
import django.utils.timezone
import nodeconductor.logging.loggers
import uuidfield.fields
import django.core.validators
import model_utils.fields
import nodeconductor.core.validators


class Migration(migrations.Migration):

    replaces = [(b'nodeconductor_sugarcrm', '0001_initial'), (b'nodeconductor_sugarcrm', '0002_add_crm_access_fields'), (b'nodeconductor_sugarcrm', '0003_paid_resources'), (b'nodeconductor_sugarcrm', '0004_add_error_message'), (b'nodeconductor_sugarcrm', '0005_crm_size'), (b'nodeconductor_sugarcrm', '0006_resource_error_message'), (b'nodeconductor_sugarcrm', '0007_init_user_limit_count_quota'), (b'nodeconductor_sugarcrm', '0008_crm_tags'), (b'nodeconductor_sugarcrm', '0009_remove_crm_size_field')]

    dependencies = [
        ('taggit', '0002_auto_20150616_2121'),
        ('structure', '0024_add_sugarcrm_to_settings'),
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='SugarCRMService',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[nodeconductor.core.validators.validate_name])),
                ('uuid', uuidfield.fields.UUIDField(unique=True, max_length=32, editable=False, blank=True)),
                ('settings', models.ForeignKey(to='structure.ServiceSettings')),
                ('customer', models.ForeignKey(to='structure.Customer')),
                ('available_for_all', models.BooleanField(default=False, help_text='Service will be automatically added to all customers projects if it is available for all')),
            ],
            options={
                'abstract': False,
                'verbose_name': 'SugarCRM service',
                'verbose_name_plural': 'SugarCRM services',
            },
            bases=(nodeconductor.core.models.SerializableAbstractMixin, nodeconductor.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='SugarCRMServiceProjectLink',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('state', django_fsm.FSMIntegerField(default=5, choices=[(0, 'New'), (5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Sync Scheduled'), (2, 'Syncing'), (3, 'In Sync'), (4, 'Erred')])),
                ('project', models.ForeignKey(to='structure.Project')),
                ('service', models.ForeignKey(to='nodeconductor_sugarcrm.SugarCRMService')),
                ('error_message', models.TextField(blank=True)),
            ],
            bases=(nodeconductor.core.models.SerializableAbstractMixin, nodeconductor.logging.loggers.LoggableMixin, models.Model),
            options={
                'verbose_name': 'SugarCRM service project link',
                'verbose_name_plural': 'SugarCRM service project links',
            },
        ),
        migrations.AlterUniqueTogether(
            name='sugarcrmserviceprojectlink',
            unique_together=set([('service', 'project')]),
        ),
        migrations.AlterUniqueTogether(
            name='sugarcrmservice',
            unique_together=set([('customer', 'settings')]),
        ),
        migrations.AddField(
            model_name='sugarcrmservice',
            name='projects',
            field=models.ManyToManyField(related_name='sugarcrm_services', through='nodeconductor_sugarcrm.SugarCRMServiceProjectLink', to=b'structure.Project'),
        ),
        migrations.CreateModel(
            name='CRM',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('description', models.CharField(max_length=500, verbose_name='description', blank=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[nodeconductor.core.validators.validate_name])),
                ('uuid', uuidfield.fields.UUIDField(unique=True, max_length=32, editable=False, blank=True)),
                ('backend_id', models.CharField(max_length=255, blank=True)),
                ('start_time', models.DateTimeField(null=True, blank=True)),
                ('state', django_fsm.FSMIntegerField(default=1, help_text='WARNING! Should not be changed manually unless you really know what you are doing.', choices=[(1, 'Provisioning Scheduled'), (2, 'Provisioning'), (3, 'Online'), (4, 'Offline'), (5, 'Starting Scheduled'), (6, 'Starting'), (7, 'Stopping Scheduled'), (8, 'Stopping'), (9, 'Erred'), (10, 'Deletion Scheduled'), (11, 'Deleting'), (13, 'Resizing Scheduled'), (14, 'Resizing'), (15, 'Restarting Scheduled'), (16, 'Restarting')])),
                ('service_project_link', models.ForeignKey(related_name='crms', on_delete=django.db.models.deletion.PROTECT, to='nodeconductor_sugarcrm.SugarCRMServiceProjectLink')),
                ('error_message', models.TextField(blank=True)),
                ('tags', taggit.managers.TaggableManager(to='taggit.Tag', through='taggit.TaggedItem', blank=True, help_text='A comma-separated list of tags.', verbose_name='Tags')),
                ('admin_username', models.CharField(max_length=60)),
                ('admin_password', models.CharField(max_length=255)),
                ('api_url', models.CharField(help_text=b'CRMs OpenStack instance URL', max_length=127)),
                ('billing_backend_id', models.CharField(help_text=b'ID of a resource in backend', max_length=255, blank=True)),
                ('last_usage_update_time', models.DateTimeField(null=True, blank=True)),
            ],
            options={
                'abstract': False,
                'verbose_name': 'CRM',
                'verbose_name_plural': 'CRMs',
            },
            bases=(nodeconductor.core.models.SerializableAbstractMixin, nodeconductor.logging.loggers.LoggableMixin, models.Model),
        ),
    ]



#!/usr/bin/env python
import sys
from setuptools import setup, find_packages


dev_requires = [
    'Sphinx==1.2.2',
]

tests_require = [
    'factory_boy==2.4.1',
    'django-celery==3.1.16',
    'mock==1.0.1',
    'mock-django==0.6.6',
    'six>=1.9.0',
]

install_requires = [
    'nodeconductor>0.102.2',
    'nodeconductor_openstack>=0.4.1',
    'sugarcrm>=0.1.1',
]

setup(
    name='nodeconductor-sugarcrm',
    version='0.2.1',
    author='OpenNode Team',
    author_email='info@opennodecloud.com',
    url='http://nodeconductor.com',
    description='NodeConductor SugarCRM adds SugarCRM support to NodeConductor',
    long_description=open('README.rst').read(),
    package_dir={'': 'src'},
    packages=find_packages('src', exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=install_requires,
    zip_safe=False,
    extras_require={
        'dev': dev_requires,
    },
    entry_points={
        'nodeconductor_extensions': (
            'nodeconductor_sugarcrm = nodeconductor_sugarcrm.extension:SugarCRMExtension',
        ),
    },
    # tests_require=tests_requires,
    include_package_data=True,
    classifiers=[
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: Apache Software License',
    ],
)



from django.contrib import admin, messages

from .backend import SugarCRMBackendError
from .models import SugarCRMServiceProjectLink, SugarCRMService, CRM
from nodeconductor.quotas.admin import QuotaInline
from nodeconductor.structure import admin as structure_admin


class CRMAdmin(structure_admin.PublishableResourceAdmin):
    actions = ['sync_quotas']
    inlines = [QuotaInline]

    def sync_quotas(self, request, queryset):
        successfully_synced = []
        for crm in queryset:
            if crm.state != CRM.States.ONLINE:
                message = 'Cannot sync quotas for CRM "%s" it is not ONLINE' % crm.name
                self.message_user(request, message, level=messages.WARNING)
                continue

            backend = crm.get_backend()
            try:
                backend.sync_user_quota()
            except SugarCRMBackendError as e:
                message = 'Cannot sync user quota for CRM "%s". Error: %s' % (crm.name, e)
                self.message_user(request, message, level=messages.ERROR)
                continue

            successfully_synced.append(crm)

        if successfully_synced:
            message = 'Quotas was successfully synced for CRMs: %s' % ', '.join(
                [crm.name for crm in successfully_synced])
            self.message_user(request, message)

    sync_quotas.short_description = "Sync quotas"


class SugarCRMServiceProjectLinkAdmin(structure_admin.ServiceProjectLinkAdmin):
    inlines = [QuotaInline]


admin.site.register(CRM, CRMAdmin)
admin.site.register(SugarCRMService, structure_admin.ServiceAdmin)
admin.site.register(SugarCRMServiceProjectLink, SugarCRMServiceProjectLinkAdmin)



import re

from rest_framework import serializers
from rest_framework.reverse import reverse

from nodeconductor.core import serializers as core_serializers
from nodeconductor.quotas import serializers as quotas_serializers
from nodeconductor.structure import serializers as structure_serializers

from . import models


class ServiceSerializer(structure_serializers.BaseServiceSerializer):

    SERVICE_ACCOUNT_FIELDS = {
        'backend_url': 'URL of template group that describes OpenStack instance provision with default parameters'
                       '(required, e.g.: http://example.com/api/template-groups/16c7675752244f5d9e870a2cb0cfeb02/)',
        'username': 'NodeConductor user username (e.g. Username)',
        'password': 'NodeConductor user password (e.g. Password)',
    }
    SERVICE_ACCOUNT_EXTRA_FIELDS = {
        'license_code': 'License code that will be used for SugarCRM activation',
        'user_data': 'User data that will be passed to CRMs OpenStack instance on creation'
                     'Word {password} will be replaced with auto-generated admin password',
        'protocol': 'CRMs access protocol',
        'phone_regex': 'RegEx for phone validation',
        'sms_email_from': 'Name of SMS email sender',
        'sms_email_rcpt': 'Name of SMS email recipient',
    }

    class Meta(structure_serializers.BaseServiceSerializer.Meta):
        model = models.SugarCRMService
        view_name = 'sugarcrm-detail'
        required_fields = 'license_code',


class ServiceProjectLinkSerializer(structure_serializers.BaseServiceProjectLinkSerializer):
    quotas = quotas_serializers.BasicQuotaSerializer(many=True, read_only=True)

    class Meta(structure_serializers.BaseServiceProjectLinkSerializer.Meta):
        model = models.SugarCRMServiceProjectLink
        view_name = 'sugarcrm-spl-detail'
        extra_kwargs = {
            'service': {'lookup_field': 'uuid', 'view_name': 'sugarcrm-detail'},
        }
        fields = structure_serializers.BaseServiceProjectLinkSerializer.Meta.fields + ('quotas',)


class CRMSerializer(structure_serializers.PublishableResourceSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='sugarcrm-detail',
        read_only=True,
        lookup_field='uuid')

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='sugarcrm-spl-detail',
        queryset=models.SugarCRMServiceProjectLink.objects.all())

    user_count = serializers.IntegerField(
        min_value=0, default=models.CRM.Quotas.user_count.default_limit, write_only=True)
    quotas = quotas_serializers.BasicQuotaSerializer(many=True, read_only=True)

    class Meta(structure_serializers.PublishableResourceSerializer.Meta):
        model = models.CRM
        view_name = 'sugarcrm-crms-detail'
        fields = structure_serializers.PublishableResourceSerializer.Meta.fields + (
            'api_url', 'user_count', 'quotas', 'instance_url')
        read_only_fields = structure_serializers.PublishableResourceSerializer.Meta.read_only_fields + (
            'api_url', 'instance_url')
        protected_fields = structure_serializers.PublishableResourceSerializer.Meta.protected_fields + ('user_count', )

    def get_fields(self):
        fields = super(CRMSerializer, self).get_fields()
        if 'request' in self.context and not self.context['request'].user.is_staff:
            fields.pop('instance_url', None)
        return fields

    def validate(self, attrs):
        if not self.instance:
            # quota check on creation. update is done through /quotas endpoint
            spl = attrs['service_project_link']
            quota = spl.quotas.get(name=models.SugarCRMServiceProjectLink.Quotas.user_limit_count)
            # usage delta for the SPL quota is equal to the limit of the resource
            delta = attrs['user_count']
            if quota.is_exceeded(delta=delta):
                raise serializers.ValidationError(
                    'Service project link users count quota is over limit (limit: %s, required: %s)' %
                    (quota.limit, quota.usage + delta))
        return attrs


class CRMUserSerializer(core_serializers.AugmentedSerializerMixin, serializers.Serializer):

    url = serializers.SerializerMethodField()
    uuid = serializers.CharField(read_only=True, source='id')
    user_name = serializers.CharField(max_length=60)
    status = serializers.CharField(max_length=30, required=False)
    last_name = serializers.CharField(max_length=30)
    first_name = serializers.CharField(max_length=30, required=False)
    email = serializers.CharField(source='email1', max_length=255, required=False)
    phone = serializers.CharField(source='phone_mobile', max_length=30, required=False, allow_blank=True)
    notify = serializers.BooleanField(write_only=True, required=False)

    class Meta(object):
        protected_fields = ['user_name', 'notify']

    def get_fields(self):
        fields = super(CRMUserSerializer, self).get_fields()
        # mark fields as not required on update
        if self.instance is not None:
            for field in fields.values():
                field.required = False
        return fields

    def get_url(self, obj):
        crm = self.context['crm']
        request = self.context['request']
        return reverse('sugarcrm-users-detail', kwargs={'crm_uuid': crm.uuid.hex, 'pk': obj.id}, request=request)

    def validate_user_name(self, value):
        crm = self.context['crm']
        users = crm.get_backend().list_users()
        exist_user_names = [u.user_name for u in users]
        if value in exist_user_names:
            raise serializers.ValidationError('User with such name already exists.')
        return value

    def validate(self, attrs):
        attrs = super(CRMUserSerializer, self).validate(attrs)
        if not self.instance:
            crm = self.context['crm']
            user_count_quota = crm.quotas.get(name=crm.Quotas.user_count)
            if user_count_quota.is_exceeded(delta=1):
                raise serializers.ValidationError(
                    'User count quota is over limit (users count: %s, max users count: %s).' %
                    (user_count_quota.usage, user_count_quota.limit))

        crm = self.context['crm']
        phone = attrs.get('phone_mobile')
        if phone:
            settings = crm.service_project_link.service.settings
            phone_regex = settings.get_option('phone_regex')
            if phone_regex and not re.search(phone_regex, phone):
                raise serializers.ValidationError({'phone': "Invalid phone number."})

        if attrs.get('notify', False) and not phone:
            raise serializers.ValidationError({'phone': "Missed phone number for notification."})

        return attrs


class UserPasswordSerializer(serializers.Serializer):
    notify = serializers.BooleanField(required=False)

    def validate(self, attrs):
        attrs = super(UserPasswordSerializer, self).validate(attrs)
        user = self.context['user']

        if attrs.get('notify', False) and not user.phone_mobile:
            raise serializers.ValidationError('User must have phone number for sending notifications.')

        return attrs



# -*- coding: utf-8

from __future__ import unicode_literals

import urllib

from django.core.management.base import BaseCommand
from django.core.urlresolvers import reverse
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType

from ...models import CRM
from nodeconductor_openstack import models as openstack_models
from nodeconductor.structure import models as structure_models, SupportedServices
from nodeconductor.template import models as template_models


class Command(BaseCommand):
    help_text = "Setup templates, service settings and services for SugarCRM default workflow"

    def handle(self, *args, **options):
        self.created_instances = []
        try:
            self.setup()
        except:
            self.rollback()
            self.stdout.write(self.style.NOTICE('  Error happened, rolling back all changes'))
            raise

    def setup(self):
        self.stdout.write(self.style.MIGRATE_HEADING('Preparation:'))
        self.base_url = (raw_input('Please enter NodeConductor base URL [http://127.0.0.1:8000]:') or
                         'http://127.0.0.1:8000')

        self.stdout.write(
            'Steps for SugarCRM setup:\n'
            ' 1. Create SugarCRM MO Customer and Project.\n'
            ' 2. Create Owner for SugarCRM MO Customer.\n'
            ' 3. Create OpenStack service settings and OpenStack service.\n'
            ' 4. Create Zabbix service settings and Zabbix service (optional).\n'
            ' 5. Create Template Group for OpenStack instance + Zabbix host.\n'
            ' 6. Create Service settings for SugarCRM.\n'
            ' 7. Create SugarCRM template.\n'
        )

        self.stdout.write(self.style.MIGRATE_HEADING('Step 1: SugarCRM MO Customer and Project'))
        # customer
        customer_name = raw_input('  Enter customer name [SugarCRM MO customer]:') or 'SugarCRM MO customer'
        customer = structure_models.Customer.objects.create(name=customer_name)
        self.created_instances.append(customer)
        customer_detail_url = self._get_admin_detail_url(customer)
        self.stdout.write('  Customer was created with given name. You can edit its details here: {}'.format(
            customer_detail_url))
        # project
        project_name = raw_input('  Enter project name [SugarCRM MO project]:') or 'SugarCRM MO project'
        project = structure_models.Project.objects.create(name=project_name, customer=customer)
        self.created_instances.append(project)
        project_detail_url = self._get_admin_detail_url(project)
        self.stdout.write('  Project was created with given name. You can edit its details here: {}'.format(
            project_detail_url))

        self.stdout.write(self.style.MIGRATE_HEADING('Step 2: Owner for SugarCRM MO Customer'))
        # user
        while True:
            username = raw_input('  Enter username of SugarCRM customer owner [SugarCRM user]:') or 'SugarCRM user'
            if get_user_model().objects.filter(username=username).exists():
                self.stdout.write('  User with such username already exists. Please choose another one.')
            else:
                break
        password = raw_input('  Enter password of SugarCRM customer owner [password]:') or 'password'
        user = get_user_model().objects.create_user(username=username, password=password)
        self.created_instances.append(user)
        customer.add_user(user, structure_models.CustomerRole.OWNER)
        user_detail_url = self._get_admin_detail_url(user)
        self.stdout.write('  Customer owner was created successfully, you can edit his details here: {}'.format(
            user_detail_url))

        self.stdout.write(self.style.MIGRATE_HEADING('Step 3: Owner for SugarCRM MO Customer'))
        # OpenStack service settings
        openstack_settings_name = (raw_input('  Enter OpenStack service settings name [SugarCRM MO OpenStack settings]:')
                                   or 'SugarCRM MO OpenStack settings')
        openstack_settings_url = self._get_admin_add_url(structure_models.ServiceSettings)
        defaults = {'customer': customer.id, 'type': 'OpenStack', 'name': openstack_settings_name, 'shared': ''}
        openstack_settings_url += '?' + urllib.urlencode(defaults)
        self.stdout.write('  Go to URL: {} and create new OpenStack service settings'.format(openstack_settings_url))
        while True:
            openstack_settings_id = raw_input('  Enter created service settings id:')
            try:
                openstack_settings = structure_models.ServiceSettings.objects.get(id=openstack_settings_id)
            except (structure_models.ServiceSettings.DoesNotExist, ValueError):
                self.stdout.write('  There is no service settings with given id, please try again.')
            else:
                self.created_instances.append(openstack_settings)
                self.stdout.write('  ServiceSettings was created successfully.')
                break
        # OpenStack service
        openstack_service_name = (raw_input('  Enter OpenStack service name [SugarCRM MO service]:')
                                  or 'SugarCRM MO service')
        openstack_service = openstack_models.OpenStackService.objects.create(
            settings=openstack_settings, name=openstack_service_name, customer=customer, available_for_all=True)
        self.created_instances[-1] = openstack_service
        self.stdout.write('  OpenStack service was created successfully.')

        if not SupportedServices.has_service_type('Zabbix'):
            self.stdout.write('Step 4: Skipped. Zabbix is not installed')
            is_zabbix_enabled = False
        else:
            from nodeconductor_zabbix import models as zabbix_models
            is_zabbix_enabled = True
            self.stdout.write(self.style.MIGRATE_HEADING('Step 4: Create Zabbix service settings and Zabbix service'))
            # Zabbix service settings
            zabbix_settings_name = (raw_input('  Enter Zabbix service settings name [SugarCRM MO Zabbix settings]:')
                                    or 'SugarCRM MO Zabbix settings')
            zabbix_settings_url = self._get_admin_add_url(structure_models.ServiceSettings)
            defaults = {'customer': customer.id, 'type': 'Zabbix', 'name': zabbix_settings_name, 'shared': ''}
            zabbix_settings_url += '?' + urllib.urlencode(defaults)
            self.stdout.write('  Go to URL: {} and create new OpenStack service settings'.format(zabbix_settings_url))
            while True:
                zabbix_settings_id = raw_input('  Enter created service settings id:')
                try:
                    zabbix_settings = structure_models.ServiceSettings.objects.get(id=zabbix_settings_id)
                except (structure_models.ServiceSettings.DoesNotExist, ValueError):
                    self.stdout.write('  There is no service settings with given id, please try again.')
                else:
                    self.created_instances.append(zabbix_settings)
                    self.stdout.write('  ServiceSettings was created successfully.')
                    break
            # Zabbix service
            zabbix_service_name = raw_input('  Enter Zabbix service name [SugarCRM MO service]:') or 'SugarCRM MO service'
            zabbix_service = zabbix_models.ZabbixService.objects.create(
                settings=zabbix_settings, name=zabbix_service_name, customer=customer, available_for_all=True)
            self.created_instances[-1] = zabbix_service
            self.stdout.write('  Zabbix service was created successfully.')

        self.stdout.write(self.style.MIGRATE_HEADING('Step 5: Template Group for OpenStack instance + Zabbix host'))
        # internal template group
        while True:
            template_group_name = (raw_input('  Enter internal template group name [SugarCRM internal]:')
                                   or 'SugarCRM internal')
            if template_models.TemplateGroup.objects.filter(name=template_group_name).exists():
                self.stdout.write(self.style.NOTICE(
                    '  Template group with such name already exists. Enter another one.'))
            else:
                break
        template_group = template_models.TemplateGroup.objects.create(name=template_group_name)
        template_group.tags.add('internal')
        self.created_instances.append(template_group)
        # OpenStack instance template
        project_url = self.base_url + reverse('project-detail', args=(project.uuid.hex,))
        service_url = self.base_url + reverse('openstack-detail', args=(openstack_service.uuid.hex,))
        instance_template = template_group.templates.create(
            order_number=1,
            resource_content_type=ContentType.objects.get_for_model(openstack_models.Instance),
            options={'project': project_url, 'service': service_url})
        self.created_instances.append(instance_template)
        template_group_detail_url = self._get_admin_detail_url(template_group)
        self.stdout.write('  Base OpenStack template group was created successfully.')
        while True:
            required_fields = 'flavor', 'image', 'system_volume_size', 'security_groups', 'data_volume_size'
            self.stdout.write('  Please go to URL {} and define {} for OpenStack instance template'.format(
                ', '.join(required_fields), template_group_detail_url))
            raw_input('  Press enter to continue')
            instance_template = template_group.templates.get(id=instance_template.id)
            instance_template_valid = True
            for field in required_fields:
                if field not in instance_template.options:
                    self.stdout.write(self.style.NOTICE(
                        '  Field {} has to be defined in OpenStack instance template'.format(field)))
                    instance_template_valid = False
            if instance_template_valid:
                break
        self.stdout.write('  OpenStack instance template was created successfully.')
        # Zabbix template
        if is_zabbix_enabled:
            service_url = self.base_url + reverse('zabbix-detail', args=(zabbix_service.uuid.hex,))
            host_template = template_group.templates.create(
                order_number=2,
                resource_content_type=ContentType.objects.get_for_model(zabbix_models.Host),
                options={'name': '{{ response.backend_id }}',
                         'visible_name': '{{ response.name }}',
                         'scope': '{{ response.url }}',
                         'project': project_url, 'service': service_url})
            self.created_instances.append(host_template)
        self.stdout.write('  Zabbix host template was created successfully.')
        self.stdout.write('  Internal template group with templates was created successfully.')

        self.stdout.write(self.style.MIGRATE_HEADING('Step 6: Service settings for SugarCRM'))
        # SugarCRM service settings
        sugarcrm_settings_name = (raw_input('  Enter SugarCRM service settings name [SugarCRM settings]:') or
                                  'SugarCRM settings')
        license_code = raw_input('  Enter SugarCRM license code:')
        protocol = raw_input('  Enter SugarCRM CRMs protocol [http]:') or 'http'
        while True:
            shared = raw_input('  Make settings shared? [Y/n]:') or 'y'
            if shared.lower() not in ('y', 'n'):
                self.stdout.write('  Please enter letter "y" or "n"')
            else:
                shared = shared.lower() == 'y'
                break
        tg_url = self.base_url + reverse('template-group-detail', args=(template_group.uuid.hex,))
        sugarcrm_settings = structure_models.ServiceSettings.objects.create(
            name=sugarcrm_settings_name,
            backend_url=tg_url,
            username=username,
            password=password,
            type='SugarCRM',
            shared=shared,
            options={'license_code': license_code, 'protocol': protocol})
        self.created_instances.append(sugarcrm_settings)
        settings_admin_url = self._get_admin_detail_url(sugarcrm_settings)
        self.stdout.write('  SugarCRM service settings was created successfully, you can edit them details here: {}'
                          .format(settings_admin_url))

        self.stdout.write(self.style.MIGRATE_HEADING('Step 7: SugarCRM template group'))
        # external template group
        while True:
            template_group_name = raw_input('  Enter SugarCRM external template group name [SugarCRM]:') or 'SugarCRM'
            if template_models.TemplateGroup.objects.filter(name=template_group_name).exists():
                self.stdout.write(self.style.NOTICE(
                    '  Template group with such name already exists. Enter another one.'))
            else:
                break
        template_group = template_models.TemplateGroup.objects.create(name=template_group_name)
        template_group.tags.add('SaaS')
        self.created_instances.append(template_group)
        # templates for template group
        settings_url = self.base_url + reverse('servicesettings-detail', args=(sugarcrm_settings.uuid.hex,))
        template = template_group.templates.create(
            order_number=1,
            resource_content_type=ContentType.objects.get_for_model(CRM),
            options={'service_settings': settings_url}
        )
        template.tags.add('IaaS')
        self.created_instances.append(template)
        template_group_admin_url = self._get_admin_detail_url(template_group)
        self.stdout.write('  External template group with templates was created successfully, you can edit its details '
                          'here: {}'.format(template_group_admin_url))
        self.stdout.write(self.style.MIGRATE_HEADING('SugarCRM setup was finished successfully.'))
        template_group_url = self.base_url + reverse('template-group-detail', args=(template_group.uuid.hex,))
        self.stdout.write('To test SugarCRM setup try to execute POST request against {} '
                          '("name" and "project" parameters are required)'.format(template_group_url + 'provision/'))

    def rollback(self):
        for instance in self.created_instances[::-1]:
            instance.delete()

    def _get_admin_detail_url(self, instance):
        name = 'admin:{app_label}_{model_name}_change'.format(
            app_label=instance._meta.app_label, model_name=instance._meta.model_name)
        return self.base_url + reverse(name, args=(instance.id,))

    def _get_admin_add_url(self, model):
        name = 'admin:{app_label}_{model_name}_add'.format(
            app_label=model._meta.app_label, model_name=model._meta.model_name)
        return self.base_url + reverse(name)
